@@ -26,6 +26,7 @@ import threading
 import time
 import uuid
 
+from amqp import exceptions as amqp_exec
 import kombu
 import kombu.connection
 import kombu.entity
@@ -90,12 +91,12 @@ rabbit_opts = [
                     '(valid only if SSL enabled).'),
     cfg.BoolOpt('heartbeat_in_pthread',
                 default=False,
-                help="EXPERIMENTAL: Run the health check heartbeat thread"
-                     "through a native python thread. By default if this"
-                     "option isn't provided the  health check heartbeat will"
-                     "inherit the execution model from the parent process. By"
-                     "example if the parent process have monkey patched the"
-                     "stdlib by using eventlet/greenlet then the heartbeat"
+                help="EXPERIMENTAL: Run the health check heartbeat thread "
+                     "through a native python thread. By default if this "
+                     "option isn't provided the  health check heartbeat will "
+                     "inherit the execution model from the parent process. By "
+                     "example if the parent process have monkey patched the "
+                     "stdlib by using eventlet/greenlet then the heartbeat "
                      "will be run through a green thread."),
     cfg.FloatOpt('kombu_reconnect_delay',
                  default=1.0,
@@ -146,7 +147,7 @@ rabbit_opts = [
                 'queue. If you just want to make sure that all queues (except '
                 'those with auto-generated names) are mirrored across all '
                 'nodes, run: '
-                """\"rabbitmqctl set_policy HA '^(?!amq\.).*' """
+                """\"rabbitmqctl set_policy HA '^(?!amq\\.).*' """
                 """'{"ha-mode": "all"}' \""""),
     cfg.IntOpt('rabbit_transient_queues_ttl',
                min=1,
@@ -168,12 +169,17 @@ rabbit_opts = [
                default=2,
                help='How often times during the heartbeat_timeout_threshold '
                'we check the heartbeat.'),
-    cfg.BoolOpt('direct_mandatory_flag',
-                default=True,
-                help='Enable/Disable the RabbitMQ mandatory flag '
-                'for direct send. The direct send is used as reply, '
-                'so the MessageUndeliverable exception is raised '
-                'in case the client queue does not exist.'),
+    cfg.IntOpt('direct_mandatory_flag',
+               default=True,
+               help='Enable/Disable the RabbitMQ mandatory flag '
+               'for direct send. The direct send is used as reply, '
+               'so the MessageUndeliverable exception is raised '
+               'in case the client queue does not exist.'),
+    cfg.BoolOpt('enable_cancel_on_failover',
+                default=False,
+                help="Enable x-cancel-on-ha-failover flag so that "
+                     "rabbitmq server will cancel and notify consumers"
+                     "when queue is down")
 ]
 
 LOG = logging.getLogger(__name__)
@@ -192,7 +198,7 @@ def _get_queue_arguments(rabbit_ha_queues, rabbit_queue_ttl):
     no longer controlled by the x-ha-policy argument when declaring a
     queue. If you just want to make sure that all queues (except those
     with auto-generated names) are mirrored across all nodes, run:
-      rabbitmqctl set_policy HA '^(?!amq\.).*' '{"ha-mode": "all"}'
+      rabbitmqctl set_policy HA '^(?!amq\\.).*' '{"ha-mode": "all"}'
 
     If the rabbit_queue_ttl option is > 0, then the queue is
     declared with the "Queue TTL" value as described here:
@@ -235,7 +241,8 @@ class Consumer(object):
 
     def __init__(self, exchange_name, queue_name, routing_key, type, durable,
                  exchange_auto_delete, queue_auto_delete, callback,
-                 nowait=False, rabbit_ha_queues=None, rabbit_queue_ttl=0):
+                 nowait=False, rabbit_ha_queues=None, rabbit_queue_ttl=0,
+                 enable_cancel_on_failover=False):
         """Init the Consumer class with the exchange_name, routing_key,
         type, durable auto_delete
         """
@@ -257,10 +264,16 @@ class Consumer(object):
             type=type,
             durable=self.durable,
             auto_delete=self.exchange_auto_delete)
+        self.enable_cancel_on_failover = enable_cancel_on_failover
 
     def declare(self, conn):
         """Re-declare the queue after a rabbit (re)connect."""
 
+        consumer_arguments = None
+        if self.enable_cancel_on_failover:
+            consumer_arguments = {
+                "x-cancel-on-ha-failover": True}
+
         self.queue = kombu.entity.Queue(
             name=self.queue_name,
             channel=conn.channel,
@@ -268,7 +281,9 @@ class Consumer(object):
             durable=self.durable,
             auto_delete=self.queue_auto_delete,
             routing_key=self.routing_key,
-            queue_arguments=self.queue_arguments)
+            queue_arguments=self.queue_arguments,
+            consumer_arguments=consumer_arguments
+        )
 
         try:
             LOG.debug('[%s] Queue.declare: %s',
@@ -325,9 +340,9 @@ class Consumer(object):
         except conn.connection.channel_errors as exc:
             # We retries once because of some races that we can
             # recover before informing the deployer
-            # bugs.launchpad.net/oslo.messaging/+bug/1581148
-            # bugs.launchpad.net/oslo.messaging/+bug/1609766
-            # bugs.launchpad.net/neutron/+bug/1318721
+            # https://bugs.launchpad.net/oslo.messaging/+bug/1581148
+            # https://bugs.launchpad.net/oslo.messaging/+bug/1609766
+            # https://bugs.launchpad.net/neutron/+bug/1318721
 
             # 406 error code relates to messages that are doubled ack'd
 
@@ -444,8 +459,6 @@ class ConnectionLock(DummyConnectionLock):
 class Connection(object):
     """Connection object."""
 
-    pools = {}
-
     def __init__(self, conf, url, purpose):
         # NOTE(viktors): Parse config options
         driver_conf = conf.oslo_messaging_rabbit
@@ -471,6 +484,7 @@ class Connection(object):
         self.kombu_failover_strategy = driver_conf.kombu_failover_strategy
         self.kombu_compression = driver_conf.kombu_compression
         self.heartbeat_in_pthread = driver_conf.heartbeat_in_pthread
+        self.enable_cancel_on_failover = driver_conf.enable_cancel_on_failover
 
         if self.heartbeat_in_pthread:
             # NOTE(hberaud): Experimental: threading module is in use to run
@@ -1014,7 +1028,8 @@ class Connection(object):
                     except (socket.timeout,
                             ConnectRefuseError,
                             OSError,
-                            kombu.exceptions.OperationalError) as exc:
+                            kombu.exceptions.OperationalError,
+                            amqp_exec.ConnectionForced) as exc:
                         LOG.info("A recoverable connection/channel error "
                                  "occurred, trying to reconnect: %s", exc)
                         self.ensure_connection()
@@ -1120,31 +1135,35 @@ class Connection(object):
         responses for call/multicall
         """
 
-        consumer = Consumer(exchange_name='',  # using default exchange
-                            queue_name=topic,
-                            routing_key='',
-                            type='direct',
-                            durable=False,
-                            exchange_auto_delete=False,
-                            queue_auto_delete=False,
-                            callback=callback,
-                            rabbit_ha_queues=self.rabbit_ha_queues,
-                            rabbit_queue_ttl=self.rabbit_transient_queues_ttl)
+        consumer = Consumer(
+            exchange_name='',  # using default exchange
+            queue_name=topic,
+            routing_key='',
+            type='direct',
+            durable=False,
+            exchange_auto_delete=False,
+            queue_auto_delete=False,
+            callback=callback,
+            rabbit_ha_queues=self.rabbit_ha_queues,
+            rabbit_queue_ttl=self.rabbit_transient_queues_ttl,
+            enable_cancel_on_failover=self.enable_cancel_on_failover)
 
         self.declare_consumer(consumer)
 
     def declare_topic_consumer(self, exchange_name, topic, callback=None,
                                queue_name=None):
         """Create a 'topic' consumer."""
-        consumer = Consumer(exchange_name=exchange_name,
-                            queue_name=queue_name or topic,
-                            routing_key=topic,
-                            type='topic',
-                            durable=self.amqp_durable_queues,
-                            exchange_auto_delete=self.amqp_auto_delete,
-                            queue_auto_delete=self.amqp_auto_delete,
-                            callback=callback,
-                            rabbit_ha_queues=self.rabbit_ha_queues)
+        consumer = Consumer(
+            exchange_name=exchange_name,
+            queue_name=queue_name or topic,
+            routing_key=topic,
+            type='topic',
+            durable=self.amqp_durable_queues,
+            exchange_auto_delete=self.amqp_auto_delete,
+            queue_auto_delete=self.amqp_auto_delete,
+            callback=callback,
+            rabbit_ha_queues=self.rabbit_ha_queues,
+            enable_cancel_on_failover=self.enable_cancel_on_failover)
 
         self.declare_consumer(consumer)
 
@@ -1155,16 +1174,18 @@ class Connection(object):
         exchange_name = '%s_fanout' % topic
         queue_name = '%s_fanout_%s' % (topic, unique)
 
-        consumer = Consumer(exchange_name=exchange_name,
-                            queue_name=queue_name,
-                            routing_key=topic,
-                            type='fanout',
-                            durable=False,
-                            exchange_auto_delete=True,
-                            queue_auto_delete=False,
-                            callback=callback,
-                            rabbit_ha_queues=self.rabbit_ha_queues,
-                            rabbit_queue_ttl=self.rabbit_transient_queues_ttl)
+        consumer = Consumer(
+            exchange_name=exchange_name,
+            queue_name=queue_name,
+            routing_key=topic,
+            type='fanout',
+            durable=False,
+            exchange_auto_delete=True,
+            queue_auto_delete=False,
+            callback=callback,
+            rabbit_ha_queues=self.rabbit_ha_queues,
+            rabbit_queue_ttl=self.rabbit_transient_queues_ttl,
+            enable_cancel_on_failover=self.enable_cancel_on_failover)
 
         self.declare_consumer(consumer)
 
@@ -1206,8 +1227,8 @@ class Connection(object):
         """Publish a message."""
 
         if not (exchange.passive or exchange.name in self._declared_exchanges):
-                exchange(self.channel).declare()
-                self._declared_exchanges.add(exchange.name)
+            exchange(self.channel).declare()
+            self._declared_exchanges.add(exchange.name)
 
         log_info = {'msg': msg,
                     'who': exchange or 'default',
@@ -1370,11 +1391,6 @@ class RabbitDriver(amqpdriver.AMQPDriverBase):
         # the pool configuration properties
         max_size = conf.oslo_messaging_rabbit.rpc_conn_pool_size
         min_size = conf.oslo_messaging_rabbit.conn_pool_min_size
-        if max_size < min_size:
-            text = 'rpc_conn_pool_size: {max_size} must be greater than' \
-                   ' or equal to conn_pool_min_size:' \
-                   ' {min_size}'.format(max_size=max_size, min_size=min_size)
-            raise RuntimeError(text)
         ttl = conf.oslo_messaging_rabbit.conn_pool_ttl
 
         connection_pool = pool.ConnectionPool(
